"""Utilities for templates."""
import re
from urllib.parse import quote_plus
from itertools import islice, accumulate
from math import copysign

from flask import url_for
from babel import numbers
from markupsafe import Markup, escape
from jinja2 import Environment, BaseLoader, pass_eval_context


def safe_display_string(string):
    """Process string for HTML display.

    If it is empty, `'<blank string>'` is shown instead.

    :param string: The string to be displayed
    :returns: Camera-ready version of the string

    """
    return re.sub('^\\s*$', '<blank string>', string)


def css_escape(name):
    """Escape string for usage as CSS id selector, as in '#{name}'.

    This basically escapes anything in ASCII range, except [_A-zA-Z0-9-] with a
    preceding backslash and keeps the rest untouched.

    See https://www.w3.org/TR/CSS22/syndata.html for further information.

    :param name: The name to be used as a CSS ID selector
    :return: Escaped string

    """
    return re.sub("(?=[\\x00-\\x7f])([^_a-zA-Z0-9-])", "\\\\\\1", name)


def parent_accounts(account):
    """Generate all parent accounts of the given account.

    :param account: GnuCash account
    :returns: Parent accounts

    """
    if account:
        yield from parent_accounts(account.parent)
        yield account


@pass_eval_context
def money(eval_ctx, amount, commodity):
    """Render monetary value for human consumption.

    :param eval_ctx: Jinja evaluation context
    :param amount: The amount to be displayed
    :param commodity: Commodity in which the amount is represented
    :returns: HTML snippet

    """
    if numbers.get_currency_symbol(commodity.mnemonic) != commodity.mnemonic:
        value = numbers.format_currency(amount, commodity.mnemonic)
    else:
        value = f"{amount} {commodity.mnemonic}"

    if eval_ctx.autoescape:
        value = escape(value)

    return Markup(
        Environment(loader=BaseLoader())
        .from_string(
            """
      <span class="text-{% if amount >= 0 %}secondary{% else %}danger{% endif %}">
        {{ value }}
      </span>
    """
        )
        .render(amount=amount, value=value)
    )


def account_url(account, *args, **kwargs):
    """Get URL to view the given account.

    Percent-encodes each account name individually (important when account name contains
    slashes) and then joins the components with slashes.

    :param account: The target account
    :returns: URL suitable for redirection ore use as hyperlink

    """
    return Markup(
        url_for(
            "book.show_account",
            account_name="/".join(
                quote_plus(account.name)
                for account in islice(
                    parent_accounts(account), 1, None
                )  # Skip root account
            ),
            *args,
            **kwargs,
        )
    )


def full_account_names(account_name):
    """Return list of full account names for all parent accounts of the given account.

    Goes up to, but not including the root account.

    Does only string operations, no actual querying of the accounts with piecash and
    is thus safe for use with non-existent accounts.

    :param account_name: Account name in question.
    :returns: List of parent account names

    """
    return accumulate(account_name.split(":"), lambda sup, sub: sup + ":" + sub)

def contra_splits(split):
    """Return list of splits on the other side of the given split in the transaction.

    In other words, return all splits whose value has the opposite sign.

    :param split: Split in question
    :returns: List of contra splits
    """
    return [contra_split for contra_split in split.transaction.splits if copysign(split.value, contra_split.value) == -float(split.value) ]

def nth(iterable, n, default=None):
    "Returns the nth item or a default value"
    return next(islice(iterable, n, None), default)



"""Functions for interacting with a GnuCash book."""
from datetime import date
from decimal import Decimal, InvalidOperation
from urllib.parse import unquote_plus, urlencode
from math import ceil

from flask import render_template, request, redirect, Blueprint
from flask import current_app as app
from piecash import Transaction, Split
from werkzeug.exceptions import BadRequest

from .auth import requires_auth, get_db_credentials
from .utils.gnucash import open_book, get_account, AccountNotFound, DatabaseLocked
from .utils.jinja import account_url

bp = Blueprint("book", __name__, url_prefix="/book")


@bp.app_errorhandler(AccountNotFound)
def handle_account_not_found(e: AccountNotFound):
    """Show error page about the desired account not being found.

    :param e: The underlying AccountNotFound exception
    :returns: Rendered HTTP Response

    """
    body = render_template("error_account_not_found.j2", account_name=e.account_name)
    return body, e.code


@bp.app_errorhandler(DatabaseLocked)
def handle_database_locked(e: DatabaseLocked):
    """Show error page about the database beeing accessed by another user.

    Includes option to repeat the causing action ignoring the lock.

    :param e: The underlying DatabaseLocked exception
    :returns: Rendered HTTP Response

    """
    # Generate path to current view, but with open_if_lock=True
    query = {"open_if_lock": "True"}
    query.update(request.args)
    url = f"{request.url}?{urlencode(query)}"

    body = render_template(
        "error_database_locked.j2",
        ignore_lock_url=url,
    )
    return body, e.code


@bp.route("/accounts/<path:account_name>")
@bp.route("/accounts/", defaults={"account_name": ""})
@requires_auth
def show_account(account_name):
    """Show the given account, including all subaccounts and transactions.

    If the account has subaccounts, a collapsible tree view of them is rendered.

    Additionally, if the account is not a placeholder, a ledger listing all
    transaction in the account is rendered, including a HTML form to add a new
    transaction.

    :param account_name: Name of the account, with / as account name separator. Each
      componnent of the account name must be urlencoded.
    :returns: Rendered HTTP Response

    """
    try:
        account_name = ":".join(unquote_plus(name) for name in account_name.split("/"))
        page = int(request.args.get('page', 1))
    except ValueError as e:
        raise BadRequest(f'Invalid query parameter: {e}') from e

    if page < 1:
        raise BadRequest(f'Invalid query parameter: page number must be positive integer: {page}')

    with open_book(
        uri_conn=app.config.DB_URI(*get_db_credentials()),
        open_if_lock=True,
        readonly=True,
    ) as book:
        account = (
            get_account(book, fullname=account_name)
            if account_name
            else book.root_account
        )

        num_pages = max(1, ceil(len(account.splits) / app.config.TRANSACTION_PAGE_LENGTH))
        if page > num_pages:
            raise BadRequest(f'Invalid query parameter: not enough pages: {page} > {num_pages}')

        return render_template(
            "account.j2",
            account=account,
            book=book,
            today=date.today(),
            num_pages=num_pages,
            page=page,
        )


@bp.route("/transaction", methods=["POST"])
@requires_auth
def add_transaction():
    """Add a new Transaction.

    All parameters are read from `request.form`.

    A positive value will transfer the desired amount from the contra account the
    receiver account ("this" account), while a negative value will deduct from the
    receiver account and credit to the contra account. Or, in other words, a negative
    value is an expense, a positive value is an income.

    :param account_name: Name of the receiver account
    :param transaction_date: Date of the transaction
    :param description: Transaction description (not split memo)
    :param value: Value of the transaction
    :param contra_account_name: Name of the contra account

    """
    try:
        account_name = request.form["account_name"]
        transaction_date = date.fromisoformat(request.form["date"])
        description = request.form["description"]
        value = Decimal(request.form["value"])
        contra_account_name = request.form["contra_account_name"]
    except (InvalidOperation, ValueError) as e:
        # TODO: Say which parameter the error is about
        raise BadRequest(f"Invalid form parameter: {e}") from e

    with open_book(
        uri_conn=app.config.DB_URI(*get_db_credentials()),
        readonly=False,
        do_backup=False,
    ) as book:
        account = get_account(book, fullname=account_name)
        contra_account = get_account(book, fullname=contra_account_name)

        if account.placeholder:
            raise BadRequest(f"{account.fullname} is a placeholder")

        if contra_account.placeholder:
            raise BadRequest(f"{contra_account.fullname} is a placeholder")

        # TODO: Support accounts with different currencies
        assert account.commodity == contra_account.commodity, (
            f"Incompatible accounts: {account.commodity} != {contra_account.commodity}."
            "Transaction form in account.j2 should not have allowed this."
        )

        common_currency = account.commodity

        # This can not fail, since currency is a valid commodity, description can be
        # any string, post_date is a valid datetime.date, account and contra_account
        # are valid non-placeholder accounts and value is a Decimal. Any other error
        # should be considered a bug.
        _ = Transaction(
            currency=common_currency,
            description=description,
            post_date=transaction_date,
            splits=[
                Split(account=account, value=value),
                Split(account=contra_account, value=-value),
            ],
        )

        book.save()

        return redirect(account_url(account))



"""App configuration management."""
import os

from flask import Config

from . import default

CONFIG_FILES = [
    "/etc/gnucash-web/config.py",
    f'{os.environ["HOME"]}/.config/gnucash-web/config.py',
]
CONFIG_ENVVAR = "GNUCASH_WEB_CONFIG"


class GnuCashWebConfig(Config):
    """GnuCash Web configuration manager.

    The main purpose of this is as opposed to the default `flask.Config` is define
    properties and helper functions for accessing the configuration options.

    """

    def __init__(self, app):
        """Initialize configuration manager for the given app.

        :param app: The Flask app
        :returns: Configuration manager

        """
        super().__init__(app.root_path)
        self.from_mapping(app.config)
        self.from_object(default)

        if CONFIG_ENVVAR in os.environ:
            if os.environ[CONFIG_ENVVAR]:
                app.logger.debug(f"Reading config file {os.environ[CONFIG_ENVVAR]}")
                self.from_envvar(CONFIG_ENVVAR)
        else:
            for path in filter(os.path.isfile, CONFIG_FILES):
                app.logger.debug(f"Reading config file {path}")
                self.from_pyfile(path)

    def DB_URI(self, user=None, password=None):
        """Get URI for the GnuCash database, possibly including credentials.

        :param user: Database username
        :param password: Database password
        :returns: Database URI for `piecash.open_book`
        :raises ValueError: If `DB_DRIVER == "sqlite"` and `user` or `password` is
          not `None`

        """
        if self["DB_DRIVER"] == "sqlite":
            if user or password:
                raise ValueError(
                    "DB_DRIVER 'sqlite' does not accept credentials"
                    " (maybe AUTH_MECHANISM 'passthrough' is set?)"
                )
            else:
                return "{DB_DRIVER}:///{DB_NAME}".format(
                    **self,
                )
        else:
            auth = ":".join(elt for elt in [user, password] if elt)
            location = "/".join(
                elt for elt in [self["DB_HOST"], self["DB_NAME"]] if elt
            )
            return "{DB_DRIVER}://{uri}".format(
                **self, uri="@".join(elt for elt in [auth, location] if elt)
            )

    @property
    def SESSION_CRYPTO_KEY(self):
        """Get key for EncryptedSession.

        We simply use the Flask SECRET_KEY here.

        :returns: 32-bit AES key as bytes

        """
        return self["SECRET_KEY"]

    def __getattr__(self, attr):
        """Entry point to get config options from gnucash_web code.

        :param attr: Name of the configuration option
        :returns: Value of the configuration option

        """
        return super().__getitem__(attr)

    def __getitem__(self, key):
        """Entry point to get config options from Flask code.

        :param key: Name of the configuration option
        :returns: Value of the configration option

        """
        if hasattr(self, key):
            return getattr(self, key)
        else:
            return super().__getitem__(key)

    def __contains__(self, key):
        """Check if configuration option is provided.

        :param key: Name of the configuration option
        :returns: True, iff the option is set

        """
        return hasattr(self, key) or super().__contains__(key)



#!python
from setuptools import setup, find_packages

setup(
    name='GnuCash Web',
    version='0.1.0-rc9',
    author='Joshua Bachmeier',
    author_email='joshua@bachmeier.cc',
    description='A simple, easy to use, mobile-friendly webinterface for GnuCash intended for self-hosting',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown; charset=UTF-8; variant=GFM',
    url='https://github.com/joshuabach/gnucash-web',
    project_urls={
        'Bug Tracker' : 'https://github.com/joshuabach/gnucash-web/issues',
        'Source Code' : 'https://github.com/joshuabach/gnucash-web',
    },
    license='GPLv3+',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'Framework :: Flask',
        'Intended Audience :: Financial and Insurance Industry',
        'Intended Audience :: Information Technology',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.8',
        'Topic :: Internet :: WWW/HTTP :: WSGI :: Application'
    ],
    keywords=['bootstrap', 'flask', 'web', 'gnucash'],

    packages=find_packages(),
    package_data={
        'gnucash_web': [
            'templates/*.j2',
            'static/*.js', 'static/*.css',
            'static/bootstrap/css/*.min.css',
            'static/bootstrap/js/*.min.js',
            'static/bootstrap-icon-font/*.css',
            'static/bootstrap-icon-font/fonts/*',
            'static/selectize/css/*.css',
            'static/selectize/js/standalone/*.js',
            'static/img/official/*/apps/*',
        ],
    },

    # Because of datetime.date.fromisoformat (3.7) and importlib.metadata (3.8)
    python_requires=">=3.8",

    install_requires=[
        'Flask>=2.0.2',
        'piecash>=1.2.0',
        'pycryptodome>=3.12.0',
        'babel>=2.9.1',
        'requests>=2.27.1',
    ],
    extras_require={
        'PostgreSQL backend': 'psycopg2',
        'MySQL / MariaDB backend': 'mysql',
    },

    entry_points={
        'console_scripts': [
            'gnucash-web = gnucash_web:cli',
        ],
    },
)



"""Default GnuCash Web configuration."""
import logging

SECRET_KEY = b'00000000000000000000000000000000'

LOG_LEVEL = logging.WARN

DB_DRIVER = 'sqlite'
DB_NAME = 'db/gnucash.sqlite'

AUTH_MECHANISM = None

TRANSACTION_PAGE_LENGTH = 25
PRESELECTED_CONTRA_ACCOUNT = None
